//  Created on : 13 Feb 2017
//  Author : M. Coroyer
//  Company : © 2017 Exoligent
//  Description:
//      This is an example of using the FipCore library.
//      The xml stations files are built from fiplabs labpro files:
//          - '2sta_worldfip_31K25.labpro'  : 31.25 Kbps
//              -> sta0_worldfip_31K25.xml
//              -> sta1_worldfip_31K25.xml
//          - '2sta_worldfip_1M.labpro'     : 1     Mbps
//              -> sta0_worldfip_1M.xml
//              -> sta1_worldfip_1M.xml
//          - '2sta_worldfip_2M5.labpro'    : 2.5   Mbps
//              -> sta0_worldfip_2M5.xml
//              -> sta1_worldfip_2M5.xml
//          - '2sta_worldfip_5M.labpro'     : 5     Mbps
//              -> sta0_worldfip_5M.xml
//              -> sta1_worldfip_5M.xml
//
//  Station0 (sta0) Configuration
//      - Arbiter: true - 1 Scan Table
//          -> Scan Table 1:
//                  SEND_ID_DAT (0x8426)
//                  SEND_ID_DAT (0x8427)
//                  SEND_ID_DAT (0x0001)
//                  SEND_MSG (5 ms)
//                  SEND_APER (10 ms)
//                  WAIT_TIME (15 ms)
//                  NEXT_MACRO ()
//      - Periodic IDs
//          -> ID(0x0001) :
//                  Type:                               Out
//                  Length:                             120 bytes
//                  Refreshment:                        true (200ms)
//                  Aperiodic's List Transfert Request: true
//                  Message Emission Channel:           false
//                  Message Reception Channel:          true
//          -> ID(0x8426) :
//                  Type:                               Out
//                  Length:                             12 bytes
//                  Refreshment:                        true (50ms)
//                  Aperiodic's List Transfert Request: false
//                  Message Emission Channel:           true (Type: Aperiodic - Channel: 0)
//                  Message Reception Channel:          false
//          -> ID(0x8427) :
//                  Type:                               In
//                  Promptness:                         true (50ms)
//                  Message Reception Channel:          false
//                  Other details: see Station1 (sta1) - ID(0x8427) 
//      - Aperiodic IDs
//          -> ID(0x8322) :
//                  Type:                               Out
//                  Length:                             10 bytes
//                  Refreshment:                        false
//                  Aperiodic's List Transfert Request: false
//                  Message Emission Channel:           false
//                  Message Reception Channel:          false
//      - Message(s)
//          -> ID (0x8426) to ID (0x8427)
//                  Source Address:                     Segment 0 - ID 0x8426
//                  Destination Address:                Segment 0 - ID 0x8427
//                  Acknowledgment:                     true
//                  Channel Emission Attached:          Type: Aperiodic - Channel: 0 (see ID(0x8426))
//
//  Station1 (sta1) Configuration
//      - Arbiter: true - 1 Scan Table
//          -> Scan Table 1:
//                  SEND_ID_DAT (0x8426)
//                  SEND_ID_DAT (0x8427)
//                  SEND_ID_DAT (0x0001)
//                  SEND_MSG (5 ms)
//                  WAIT_TIME (5 ms)
//                  NEXT_MACRO ()
//      - Periodic IDs
//          -> ID(0x8426) :
//                  Type:                               In
//                  Promptness:                         true (50ms)
//                  Message Reception Channel:          false
//                  Other details: see Station0 (sta0) - ID(0x8426) 
//          -> ID(0x8427) :
//                  Type:                               Out
//                  Length:                             12 bytes
//                  Refreshment:                        true (200 ms)
//                  Aperiodic's List Transfert Request: false
//                  Message Emission Channel:           false
//                  Message Reception Channel:          true
//      - Aperiodic IDs
//          -> ID(0x8322) :
//                  Type:                               In
//                  Promptness:                         true (50ms)
//                  Message Reception Channel:          false
//                  Other details: see Station0 (sta0) - ID(0x8322)
//
//  In this example:
//      station0_process routine:
//          Open Device / Load Configuration / Run Bus Arbiter / Run OPC UA Server (Blocking mode)
//          To Quit, the user has to enter ctrl-c
//
//      station1_process routine:
//          Open Device / Load Configuration / Run Bus Arbiter / Run OPC UA Server (Blocking mode)
//          To Quit, the user has to enter ctrl-c
//
//  Note: In the example, ID(0x0001) is only used to treat the aperiodic's list requests; that's why it is
//  present in periodic window in scan table

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>

#include "fipcore.h"

#if defined(_WIN32) || defined(_WIN64)
#   include <windows.h>
#   include <conio.h>
#elif defined(__linux__) || defined(__APPLE__) || defined(__QNX__)
#   include <unistd.h>     // usleep()
#   include <termios.h>
#   include <sys/types.h>
#   include <sys/time.h>

void changemode(int dir)
{
    static struct termios oldt, newt;
    
    if (dir == 1) {
        tcgetattr(STDIN_FILENO, &oldt);
        newt = oldt;
        newt.c_lflag &= ~(ICANON | ECHO);
        tcsetattr(STDIN_FILENO, TCSANOW, &newt);
    } else
        tcsetattr(STDIN_FILENO, TCSANOW, &oldt);
}

int kbhit(void)
{
    struct timeval tv;
    fd_set rdfs;
    
    tv.tv_sec = 0;
    tv.tv_usec = 0;
    
    FD_ZERO(&rdfs);
    FD_SET(STDIN_FILENO, &rdfs);
    
    select(STDIN_FILENO + 1, &rdfs, NULL, NULL, &tv);
    return FD_ISSET(STDIN_FILENO, &rdfs);
}
#endif


uint16_t event_parameter = 0;
bool     opc_is_running = true;

char* sta_xml_path[2][4] = {
#if defined(_TST_FLAT_PATH)
    {"./sta0_worldfip_31K25.xml", "./sta0_worldfip_1M.xml", "./sta0_worldfip_2M5.xml", "./sta0_worldfip_5M.xml"},
    {"./sta1_worldfip_31K25.xml", "./sta1_worldfip_1M.xml", "./sta1_worldfip_2M5.xml", "./sta1_worldfip_5M.xml"}
#else
    {"../xml/sta0_worldfip_31K25.xml", "../xml/sta0_worldfip_1M.xml", "../xml/sta0_worldfip_2M5.xml", "../xml/sta0_worldfip_5M.xml"},
    {"../xml/sta1_worldfip_31K25.xml", "../xml/sta1_worldfip_1M.xml", "../xml/sta1_worldfip_2M5.xml", "../xml/sta1_worldfip_5M.xml"}
#endif
};

void station0_process(uint8_t device_type, uint8_t device_index, char* xml_path);
void station1_process(uint8_t device_type, uint8_t device_index, char* xml_path);
void print_device_info(FIP_DEVICECONF* device_conf);
void print_conf(const FIP_USERCONF* fuc);

static void stopHandler(int sign) {
    opc_is_running = false;
}

int main(int argCount, char *argValue[])
{
    uint32_t ret = SUCCESS;
    uint8_t i = 0;
    uint8_t device_type;
    uint8_t device_index;
    uint8_t station_number;
    uint8_t station_speed;
    FIP_DEVICECONF devices_list[20];
    uint8_t nb_detected_devices = 0;
    bool    is_matched = false;
    char    temp[10];

    // Device Enumeration
    printf("/*****************************************************************************/\n");
    printf("/*                          DEVICES ON COMPUTER                              */\n");
    printf("/*****************************************************************************/\n");

    ret = fipcore_get_devices_number(&nb_detected_devices);
    if (ret != SUCCESS) {
        printf("fipcore_get_devices_number - %s\n", fipcore_get_error_str(ret));
        return 0;
    }

    printf("NUMBER OF DETECTED DEVICES - %d\n", nb_detected_devices);

    ret = fipcore_get_devices_list(devices_list, &nb_detected_devices);
    if (ret != SUCCESS) {
        printf("fipcore_get_devices_list - %s\n", fipcore_get_error_str(ret));
        return 0;
    }

    for (i = 0; i < nb_detected_devices; i++)
        print_device_info(&(devices_list[i]));

    if (argCount < 4) {
        printf("This application must have this form: fipcore_opcua_blocking.exe 'device_type' 'device_index' 'station_number' 'station_speed' \n");
        printf("'device_type': 1: PCI_PCIE_TYPE - 2: USB_TYPE\n");
        printf("'device_index': 0 to 255\n");
        printf("'station_number': 0: Station 0 - 1: Station 1\n");
        printf("'station_speed': 0: 31.25 Kbps - 1: 1 Mbps - 2: 2.5 Mbps - 3: 5Mbps\n");
        return 0;
    }

    sscanf (argValue[1], "%s", temp);
    device_type = (uint8_t) atoi(temp);
    memset(temp, 0, sizeof(temp));

    sscanf (argValue[2], "%s", temp);
    device_index = (uint8_t) atoi(temp);
    memset(temp, 0, sizeof(temp));

    sscanf (argValue[3], "%s", temp);
    station_number = (uint8_t) atoi(temp);
    memset(temp, 0, sizeof(temp));

    sscanf (argValue[4], "%s", temp);
    station_speed = (uint8_t) atoi(temp);
    memset(temp, 0, sizeof(temp));
    
    if (!(device_type < MAX_TYPE)) {
        printf("Invalid 'device_type' parameter.\n");
        return 0;
    }
    if (station_number > 1) {
        printf("Invalid 'station_number' parameter.\n");
        return 0; 
    }
    if (station_speed > 3) {
        printf("Invalid 'station_speed' parameter.\n");
        return 0; 
    }

    for (i = 0; i < nb_detected_devices; i++) {

        if ((devices_list[i].device_type == device_type) && 
            (devices_list[i].device_index == device_index)) {

                if (devices_list[i].fip_speed != station_speed) {

                    if (!((devices_list[i].fip_speed == SPEED_2M5) && 
                        (station_speed == SPEED_5M))) {
                        // Support 2.5 Mbps device with 5 Mbps configuration

                        // else exit
                        printf("Station speed must match with device speed.\n");
                        return 0;
                    }
                }
                is_matched = true;
        }
    }

    if (is_matched) {
        if (station_number == 0)
            station0_process(device_type, device_index, sta_xml_path[station_number][station_speed]);
        else
            station1_process(device_type, device_index, sta_xml_path[station_number][station_speed]);
    } else
        printf("No device was found with device_type (%d) and device_index (%d).\n", device_type, device_index);

    return 0;
}

void station0_process(uint8_t device_type, uint8_t device_index, char* xml_path)
{
    signal(SIGINT, stopHandler); /* catches ctrl-c */

    uint32_t ret = SUCCESS;
    //FIP_USERCONF user_conf = {0};
    PRODUCED_VARIABLE p_var = {0};
    fipcore_device_handle* sta0_hnd = NULL;

    printf("/*****************************************************************************/\n");
    printf("/*                          STATION 0 TEST                                   */\n");
    printf("/*****************************************************************************/\n");

    sta0_hnd = fipcore_device_open(device_type, device_index, &ret);
    if (ret != SUCCESS) {
        printf("STATION 0 - %s\n", fipcore_get_error_str(ret));
        goto _exit1;
    }

    ret = fipcore_init_network(sta0_hnd, XML, xml_path, NULL);
    if (ret != SUCCESS)
        goto _exit;

    // Optional Configuration Information - 
    // (To execute after fipcore_init_network function)
    // fipcore_get_config(sta0_hnd, &user_conf);
    // print_conf(&user_conf);

    ret = fipcore_start_network(sta0_hnd);
    if (ret != SUCCESS)
        goto _exit;

    ret = fipcore_start_ba(sta0_hnd, 1);
    if (ret != SUCCESS)
        goto _exit;

    // Output variables init
    fipcore_write_var_by_id(sta0_hnd, 0x0001, &p_var);
    fipcore_write_var_by_id(sta0_hnd, 0x8426, &p_var);
    fipcore_write_var_by_id(sta0_hnd, 0x8322, &p_var);

#if defined(__linux__) || defined(__APPLE__)
    changemode(1);
#endif
    printf("-> fipcore_opcua_run - [Press ctrl-c to quit]\n");
    // OPC UA START SERVER
    ret = fipcore_opcua_run(sta0_hnd, 16664, &opc_is_running, NULL);
    if (ret != SUCCESS)
        goto _exit;

#if defined(__linux__) || defined(__APPLE__)
    changemode(0);
#endif
    
    ret = fipcore_stop_ba(sta0_hnd);

_exit:  
    if (ret != SUCCESS)
        printf("STATION 0 - %s\n", fipcore_get_error_str(ret));

    ret = fipcore_stop_network(sta0_hnd);
    if (ret != SUCCESS)
        printf("STATION 0 - %s\n", fipcore_get_error_str(ret));
_exit1:
    ret = fipcore_device_close(sta0_hnd);
    if (ret != SUCCESS)
        printf("STATION 0 - %s\n", fipcore_get_error_str(ret));
}

void station1_process(uint8_t device_type, uint8_t device_index, char* xml_path)
{
    signal(SIGINT, stopHandler); /* catches ctrl-c */
    uint32_t ret = SUCCESS;
    //FIP_USERCONF user_conf = {0};
    PRODUCED_VARIABLE p_var = {0};
    fipcore_device_handle* sta1_hnd = NULL;

    printf("/*****************************************************************************/\n");
    printf("/*                          STATION 1 TEST                                   */\n");
    printf("/*****************************************************************************/\n");

    sta1_hnd = fipcore_device_open(device_type, device_index, &ret);
    if (ret != SUCCESS) {
        printf("STATION 1 - %s\n", fipcore_get_error_str(ret));
        goto _exit1;
    }

    ret = fipcore_init_network(sta1_hnd, XML, xml_path, NULL);
    if (ret != SUCCESS)
        goto _exit;

    // Optional Configuration Information - 
    // (To execute after fipcore_init_network function)
    //fipcore_get_config(sta1_hnd, &user_conf);
    //print_conf(&user_conf);

    ret = fipcore_start_network(sta1_hnd);
    if (ret != SUCCESS)
        goto _exit;

    ret = fipcore_start_ba(sta1_hnd, 1);
    if (ret != SUCCESS)
        goto _exit;

    // Output variables init
    fipcore_write_var_by_id(sta1_hnd, 0x8427, &p_var);

#if defined(__linux__) || defined(__APPLE__)
    changemode(1);
#endif
    printf("-> fipcore_opcua_run - [Press ctrl-c to quit]\n");
    // OPC UA START SERVER
    ret = fipcore_opcua_run(sta1_hnd, 16664, &opc_is_running, NULL);
    if (ret != SUCCESS)
        goto _exit;

#if defined(__linux__) || defined(__APPLE__)
    changemode(0);
#endif
    
    ret = fipcore_stop_ba(sta1_hnd);

_exit:  
    if (ret != SUCCESS)
        printf("STATION 1 - %s\n", fipcore_get_error_str(ret));

    ret = fipcore_stop_network(sta1_hnd);
    if (ret != SUCCESS)
        printf("STATION 1 - %s\n", fipcore_get_error_str(ret));
_exit1:
    ret = fipcore_device_close(sta1_hnd);
    if (ret != SUCCESS)
        printf("STATION 1 - %s\n", fipcore_get_error_str(ret));
}

void print_device_info(FIP_DEVICECONF* device_conf)
{
    switch(device_conf->device_type) {
    case PCI_PCIE_TYPE:     printf("   device_type     : PCI_PCIE_TYPE \n"); break;
    case USB_TYPE:          printf("   device_type     : USB_TYPE \n"); break;
    default:                printf("   device_type     : UNKNOWN_TYPE \n"); break;
    }
    printf("   device_index    : %d \n", device_conf->device_index);
    printf("   product_number  : 0x%08x \n", device_conf->product_number);
    printf("   device_version  : %d.%d.%d \n", device_conf->device_version.major,
                                            device_conf->device_version.minor,
                                            device_conf->device_version.revision);
    switch(device_conf->fip_speed) {
    case SPEED_31K25:   printf("   fip_speed       : SPEED_31K25 \n"); break;
    case SPEED_1M:      printf("   fip_speed       : SPEED_1M \n"); break;
    case SPEED_2M5:     printf("   fip_speed       : SPEED_2M5 or SPEED_5M \n"); break;
    default:            printf("   fip_speed       : UNKNOWN_SPEED \n"); break;
    }
    switch(device_conf->fip_impedance) {
    case IMPEDANCE_150: printf("   fip_impedance   : IMPEDANCE_150 \n\n"); break;
    case IMPEDANCE_120: printf("   fip_impedance   : IMPEDANCE_120 \n\n"); break;
    default:            printf("   fip_impedance   : UNKNOWN_IMPEDANCE \n\n"); break;
    }
}

void print_conf(const FIP_USERCONF* fuc)
{
    uint16_t cp = 0;
    uint16_t i  = 0;


    printf("/*****************************************************************************/\n");
    printf("/*                              FIP_USERCONF                                 */\n");
    printf("/*****************************************************************************/\n");
    if (fuc != NULL) {
        printf("    - this_station               : \n");
        printf("        - subscriber_number      : %d \n", fuc->this_station->subscriber_number);
        printf("        - subscriber_name        : %s \n", fuc->this_station->subscriber_name);
        printf("    - speed                      : %d \n", fuc->speed);
        printf("    - t_slot                     : %d \n", fuc->t_slot);
        printf("    - silence                    : %d \n", fuc->silence);
        printf("    - turn_around                : %d \n", fuc->turn_around);
        printf("    - frame_type                 : %d \n", fuc->frame_type);
        printf("    - is_arbiter                 : %d \n", fuc->is_arbiter);
        printf("    - arbiter                    : \n");
        printf("        - ba_max_subscriber_number : %d \n", fuc->arbiter->ba_max_subscriber_number);
        printf("        - ba_priority            : %d \n", fuc->arbiter->ba_priority);
        printf("        - ba_max_priority        : %d \n", fuc->arbiter->ba_max_priority);
        printf("        - ba_optimization        : %d \n", fuc->arbiter->ba_optimization);
        printf("        - nb_ba_1_instructions   : %d \n", cp = fuc->arbiter->nb_ba_1_instructions);
        printf("        - ba_1_instructions      : \n");
        {
            INSTRUCTION* instructions = fuc->arbiter->ba_1_instructions;
            for (i = 0; i < cp; i++)
                printf("            - %s \n", (instructions + i)->op_code);
        }
        printf("        - nb_ba_2_instructions   : %d \n", cp = fuc->arbiter->nb_ba_2_instructions);
        printf("        - ba_2_instructions      : \n");
        {
            INSTRUCTION* instructions = fuc->arbiter->ba_2_instructions;
            for (i = 0; i < cp; i++)
                printf("            - %s \n", (instructions + i)->op_code);
        }   
        printf("    - nb_variables               : %d \n", cp = fuc->nb_variables);
        printf("    - variables                  : \n");
        {
            for (i = 0; i < cp; i++) {
                printf("        - number %d \n", i);
                printf("            - name               : %s \n", fuc->variables[i].name);
                printf("            - identifier         : 0x%04x \n", fuc->variables[i].identifier);
                printf("            - pdu                : 0x%04x \n", fuc->variables[i].pdu);
                printf("            - length             : %d \n", fuc->variables[i].length);
                printf("            - type               : %d \n", fuc->variables[i].type);
                printf("            - refreshment_type   : %d \n", fuc->variables[i].refreshment_type);
                printf("            - refreshment_period : 0x%04x \n", fuc->variables[i].refreshment_period);
                printf("            - promptitude_type   : %d \n", fuc->variables[i].promptitude_type);
                printf("            - promptitude_period : 0x%04x \n", fuc->variables[i].promptitude_period);
                printf("            - msg_channel        : %d \n", fuc->variables[i].msg_channel);
                printf("            - messages_service   : 0x%02x \n", fuc->variables[i].messages_service);
            }
        }
        printf("    - nb_messages                : %d \n", cp = fuc->nb_messages);
        printf("    - messages                   : \n");
        {
            for (i = 0; i < cp; i++) {
                printf("        - number %d \n", i);
                printf("            - name               : %s \n", fuc->messages[i].name);
                printf("            - type               : %d \n", fuc->messages[i].type);
                printf("            - channel            : %d \n", fuc->messages[i].channel);
                printf("            - acknowledgment     : %d \n", fuc->messages[i].acknowledgment);
                printf("            - destination_identifier: 0x%04x  \n", fuc->messages[i].destination_identifier);
                printf("            - destination_segment: 0x%02x  \n", fuc->messages[i].destination_segment);
                printf("            - source_identifier  : 0x%04x  \n", fuc->messages[i].source_identifier);
                printf("            - source_segment     : 0x%02x  \n", fuc->messages[i].source_segment);
            }
        }
        printf("     - nb_tags            : %d \n", cp = fuc->nb_tags);
        printf("     - tags               : \n");
        {
            for (i = 0; i < cp; i++) {
                printf("        - number %d \n", i);
                printf("            - tag_name      : %s \n", fuc->tags[i].tag_name);
                printf("            - tag_id        : %d \n", fuc->tags[i].tag_id);
                printf("            - type_name     : %s \n", fuc->tags[i].type_name);
                printf("            - type_id       : %d \n", fuc->tags[i].type_id);
                printf("            - access_rights : %d \n", fuc->tags[i].access_rights);
                printf("            - description   : %s \n", fuc->tags[i].description);
                printf("            - identifier    : 0x%04x \n", fuc->tags[i].identifier);
                printf("            - start_bit     : %d \n", fuc->tags[i].start_bit);
                printf("            - stop_bit      : %d \n", fuc->tags[i].stop_bit);
                printf("            - endianness    : %d \n", fuc->tags[i].endianness);
            }
        }
    } else {
        printf("    - user_conf               : \n");
        printf("        - unallocated \n");
    }
}